##
# File: ChemCompUtil.py
# Date: 18-Feb-2022
#
# Updates:
# 27-Apr-2022  M. Yokochi - enable to use cached data for standard residues
# 11-Nov-2022  M. Yokochi - add getProtonsInSameGroup() (NMR restraint remediation)
# 13-Jun-2023  M. Yokochi - add getEffectiveFormulaWeight()
# 07-Dec-2023  M. Yokochi - add support for PTM items (backbone, n_terminal, c_terminal atom flags)
# 13-Dec-2023  M. Yokochi - add getAtomsBasedOnGreekLetterSystem(), peptideLike() and getTypeOfCompId() (DAOTHER-8945)
# 19-Apr-2024  M. Yokochi - add getRepMethyleneOrAminoProtons() and getNonRepMethyleneOrAminoProtons() (DAOTHER-9317)
##
""" Wrapper class for retrieving chemical component dictionary.
    @author: Masashi Yokochi
"""
import os
import sys
import pickle

from rmsd.calculate_rmsd import NAMES_ELEMENT, ELEMENT_WEIGHTS  # noqa: F401 pylint: disable=no-name-in-module, import-error, unused-import


try:
    from wwpdb.utils.config.ConfigInfo import getSiteId
    from wwpdb.utils.config.ConfigInfoApp import ConfigInfoAppCc
    from wwpdb.utils.nmr.io.ChemCompIo import ChemCompReader
    from wwpdb.utils.nmr.AlignUtil import (emptyValue,
                                           monDict3,
                                           protonBeginCode)
    cICc = ConfigInfoAppCc(getSiteId())
    CC_CVS_PATH = cICc.get_site_cc_cvs_path()
except ImportError:
    from nmr.io.ChemCompIo import ChemCompReader
    from nmr.AlignUtil import (emptyValue,
                               monDict3,
                               protonBeginCode)
    CC_CVS_PATH = os.path.dirname(__file__) + '/ligand_dict'  # need to setup 'ligand_dict' CCD resource for NMR restraint processing


class ChemCompUtil:
    """ Wrapper class for retrieving chemical component dictionary.
    """

    def __init__(self, verbose=False, log=sys.stderr):

        # pickle file name of cached dictionary for standard residues
        self.__cacheFile = os.path.dirname(__file__) + '/chem_comp_util/std_chem_comp.pkl'

        self.__ccR = ChemCompReader(verbose, log)
        self.__ccR.setCachePath(CC_CVS_PATH)

        self.lastCompId = None
        self.lastStatus = False
        self.lastChemCompDict = None
        self.lastAtomList = None
        self.lastBonds = None

        # taken from wwpdb.utils.nmr.io.ChemCompIo
        _chemCompAtomDict = [
            ('_chem_comp_atom.comp_id', '%s', 'str', ''),
            ('_chem_comp_atom.atom_id', '%s', 'str', ''),
            ('_chem_comp_atom.alt_atom_id', '%s', 'str', ''),
            ('_chem_comp_atom.type_symbol', '%s', 'str', ''),
            ('_chem_comp_atom.charge', '%s', 'str', ''),
            ('_chem_comp_atom.pdbx_align', '%s', 'str', ''),
            ('_chem_comp_atom.pdbx_aromatic_flag', '%s', 'str', ''),
            ('_chem_comp_atom.pdbx_leaving_atom_flag', '%s', 'str', ''),
            ('_chem_comp_atom.pdbx_stereo_config', '%s', 'str', ''),
            ('_chem_comp_atom.model_Cartn_x', '%s', 'str', ''),
            ('_chem_comp_atom.model_Cartn_y', '%s', 'str', ''),
            ('_chem_comp_atom.model_Cartn_z', '%s', 'str', ''),
            ('_chem_comp_atom.pdbx_model_Cartn_x_ideal', '%s', 'str', ''),
            ('_chem_comp_atom.pdbx_model_Cartn_y_ideal', '%s', 'str', ''),
            ('_chem_comp_atom.pdbx_model_Cartn_z_ideal', '%s', 'str', ''),
            ('_chem_comp_atom.pdbx_component_atom_id', '%s', 'str', ''),
            ('_chem_comp_atom.pdbx_component_comp_id', '%s', 'str', ''),
            ('_chem_comp_atom.pdbx_ordinal', '%s', 'str', ' '),
            ('_chem_comp_atom.pdbx_backbone_atom_flag', '%s', 'str', ''),
            ('_chem_comp_atom.pdbx_n_terminal_atom_flag', '%s', 'str', ''),
            ('_chem_comp_atom.pdbx_c_terminal_atom_flag', '%s', 'str', '')
        ]

        atomId = next(d for d in _chemCompAtomDict if d[0] == '_chem_comp_atom.atom_id')
        self.ccaAtomId = _chemCompAtomDict.index(atomId)

        altAtomId = next(d for d in _chemCompAtomDict if d[0] == '_chem_comp_atom.alt_atom_id')
        self.ccaAltAtomId = _chemCompAtomDict.index(altAtomId)

        aromaticFlag = next(d for d in _chemCompAtomDict if d[0] == '_chem_comp_atom.pdbx_aromatic_flag')
        self.ccaAromaticFlag = _chemCompAtomDict.index(aromaticFlag)

        leavingAtomFlag = next(d for d in _chemCompAtomDict if d[0] == '_chem_comp_atom.pdbx_leaving_atom_flag')
        self.ccaLeavingAtomFlag = _chemCompAtomDict.index(leavingAtomFlag)

        typeSymbol = next(d for d in _chemCompAtomDict if d[0] == '_chem_comp_atom.type_symbol')
        self.ccaTypeSymbol = _chemCompAtomDict.index(typeSymbol)

        cartnX = next(d for d in _chemCompAtomDict if d[0] == '_chem_comp_atom.model_Cartn_x')
        cartnY = next(d for d in _chemCompAtomDict if d[0] == '_chem_comp_atom.model_Cartn_y')
        cartnZ = next(d for d in _chemCompAtomDict if d[0] == '_chem_comp_atom.model_Cartn_z')
        self.ccaCartnX = _chemCompAtomDict.index(cartnX)
        self.ccaCartnY = _chemCompAtomDict.index(cartnY)
        self.ccaCartnZ = _chemCompAtomDict.index(cartnZ)

        backboneAtomFlag = next(d for d in _chemCompAtomDict if d[0] == '_chem_comp_atom.pdbx_backbone_atom_flag')
        self.ccaBackboneAtomFlag = _chemCompAtomDict.index(backboneAtomFlag)

        nTerminalAtomFlag = next(d for d in _chemCompAtomDict if d[0] == '_chem_comp_atom.pdbx_n_terminal_atom_flag')
        self.ccaNTerminalAtomFlag = _chemCompAtomDict.index(nTerminalAtomFlag)

        cTerminalAtomFlag = next(d for d in _chemCompAtomDict if d[0] == '_chem_comp_atom.pdbx_c_terminal_atom_flag')
        self.ccaCTerminalAtomFlag = _chemCompAtomDict.index(cTerminalAtomFlag)

        # taken from wwpdb.utils.nmr.io.ChemCompIo
        _chemCompBondDict = [
            ('_chem_comp_bond.comp_id', '%s', 'str', ''),
            ('_chem_comp_bond.atom_id_1', '%s', 'str', ''),
            ('_chem_comp_bond.atom_id_2', '%s', 'str', ''),
            ('_chem_comp_bond.value_order', '%s', 'str', ''),
            ('_chem_comp_bond.pdbx_aromatic_flag', '%s', 'str', ''),
            ('_chem_comp_bond.pdbx_stereo_config', '%s', 'str', ''),
            ('_chem_comp_bond.pdbx_ordinal', '%s', 'str', '')
        ]

        atomId1 = next(d for d in _chemCompBondDict if d[0] == '_chem_comp_bond.atom_id_1')
        self.ccbAtomId1 = _chemCompBondDict.index(atomId1)

        atomId2 = next(d for d in _chemCompBondDict if d[0] == '_chem_comp_bond.atom_id_2')
        self.ccbAtomId2 = _chemCompBondDict.index(atomId2)

        aromaticFlag = next(d for d in _chemCompBondDict if d[0] == '_chem_comp_bond.pdbx_aromatic_flag')
        self.ccbAromaticFlag = _chemCompBondDict.index(aromaticFlag)

        valueOrder = next(d for d in _chemCompBondDict if d[0] == '_chem_comp_bond.value_order')
        self.ccbValueOrder = _chemCompBondDict.index(valueOrder)

        def load_dict_from_pickle(file_name):
            """ Load cached dictionary from pickle file.
            """

            if os.path.exists(file_name):

                with open(file_name, 'rb') as ifh:
                    return pickle.load(ifh)

            return {}

        self.__cachedDict = load_dict_from_pickle(self.__cacheFile)
        self.__failedCompId = []

    def updateChemCompDict(self, compId):
        """ Update CCD information for a given comp_id.
            @return: True for successfully update CCD information or False for the case a given comp_id does not exist in CCD
        """

        if compId in emptyValue:
            return False

        compId = compId.upper()

        if compId in self.__failedCompId:
            return False

        if compId != self.lastCompId:
            self.lastStatus = False if '_' in compId else self.__ccR.setCompId(compId)
            self.lastCompId = compId

            if self.lastStatus:
                if compId in self.__cachedDict:
                    self.lastChemCompDict = self.__cachedDict[compId]['chem_comp']
                    self.lastAtomList = self.__cachedDict[compId]['chem_comp_atom']
                    self.lastBonds = self.__cachedDict[compId]['chem_comp_bond']
                else:
                    self.lastChemCompDict = self.__ccR.getChemCompDict()
                    self.lastAtomList = self.__ccR.getAtomList()
                    self.lastBonds = self.__ccR.getBonds()
                    self.__cachedDict[compId] = {'chem_comp': self.lastChemCompDict,
                                                 'chem_comp_atom': self.lastAtomList,
                                                 'chem_comp_bond': self.lastBonds}

            else:
                self.__failedCompId.append(compId)

        return self.lastStatus

    def getMethylAtoms(self, compId):
        """ Return atoms in methyl group of a given comp_id.
        """

        if compId != self.lastCompId and not self.updateChemCompDict(compId):
            return []

        atmList = []

        carbons = (a[self.ccaAtomId] for a in self.lastAtomList if a[self.ccaTypeSymbol] == 'C')

        for carbon in carbons:
            protons = [(b[self.ccbAtomId1] if b[self.ccbAtomId1] != carbon else b[self.ccbAtomId2])
                       for b in self.lastBonds
                       if (b[self.ccbAtomId1] == carbon and b[self.ccbAtomId2][0] in protonBeginCode)
                       or (b[self.ccbAtomId2] == carbon and b[self.ccbAtomId1][0] in protonBeginCode)]
            if len(protons) != 3:
                continue
            atmList.append(carbon)
            atmList.extend(protons)

        return atmList

    def getRepMethylProtons(self, compId):
        """ Return representative protons in methyl group of a given comp_id.
        """

        if compId != self.lastCompId and not self.updateChemCompDict(compId):
            return []

        atmList = []

        carbons = (a[self.ccaAtomId] for a in self.lastAtomList if a[self.ccaTypeSymbol] == 'C')

        for carbon in carbons:
            protons = [(b[self.ccbAtomId1] if b[self.ccbAtomId1] != carbon else b[self.ccbAtomId2])
                       for b in self.lastBonds
                       if (b[self.ccbAtomId1] == carbon and b[self.ccbAtomId2][0] in protonBeginCode)
                       or (b[self.ccbAtomId2] == carbon and b[self.ccbAtomId1][0] in protonBeginCode)]
            if len(protons) != 3:
                continue
            atmList.append(protons[0])

        return atmList

    def getNonRepMethylProtons(self, compId):
        """ Return non-representative protons in methyl group of a given comp_id.
        """

        if compId != self.lastCompId and not self.updateChemCompDict(compId):
            return []

        atmList = []

        carbons = (a[self.ccaAtomId] for a in self.lastAtomList if a[self.ccaTypeSymbol] == 'C')

        for carbon in carbons:
            protons = [(b[self.ccbAtomId1] if b[self.ccbAtomId1] != carbon else b[self.ccbAtomId2])
                       for b in self.lastBonds
                       if (b[self.ccbAtomId1] == carbon and b[self.ccbAtomId2][0] in protonBeginCode)
                       or (b[self.ccbAtomId2] == carbon and b[self.ccbAtomId1][0] in protonBeginCode)]
            if len(protons) != 3:
                continue
            atmList.extend(protons[1:])

        return atmList

    def getRepMethyleneOrAminoProtons(self, compId):
        """ Return representative protons in methylene/amino group of a given comp_id.
        """

        if compId != self.lastCompId and not self.updateChemCompDict(compId):
            return []

        atmList = []

        corns = (a[self.ccaAtomId] for a in self.lastAtomList if a[self.ccaTypeSymbol] in ('C', 'N'))

        for corn in corns:
            protons = [(b[self.ccbAtomId1] if b[self.ccbAtomId1] != corn else b[self.ccbAtomId2])
                       for b in self.lastBonds
                       if (b[self.ccbAtomId1] == corn and b[self.ccbAtomId2][0] in protonBeginCode)
                       or (b[self.ccbAtomId2] == corn and b[self.ccbAtomId1][0] in protonBeginCode)]
            if len(protons) != 2:
                continue
            atmList.append(protons[0])

        return atmList

    def getNonRepMethyleneOrAminoProtons(self, compId):
        """ Return non-representative protons in methylene/amino group of a given comp_id.
        """

        if compId != self.lastCompId and not self.updateChemCompDict(compId):
            return []

        atmList = []

        corns = (a[self.ccaAtomId] for a in self.lastAtomList if a[self.ccaTypeSymbol] in ('C', 'N'))

        for corn in corns:
            protons = [(b[self.ccbAtomId1] if b[self.ccbAtomId1] != corn else b[self.ccbAtomId2])
                       for b in self.lastBonds
                       if (b[self.ccbAtomId1] == corn and b[self.ccbAtomId2][0] in protonBeginCode)
                       or (b[self.ccbAtomId2] == corn and b[self.ccbAtomId1][0] in protonBeginCode)]
            if len(protons) != 2:
                continue
            atmList.extend(protons[1:])

        return atmList

    def getRepAminoProtons(self, compId):
        """ Return representative protons in amino group of a given comp_id.
        """

        if compId != self.lastCompId and not self.updateChemCompDict(compId):
            return []

        atmList = []

        ns = (a[self.ccaAtomId] for a in self.lastAtomList if a[self.ccaTypeSymbol] == 'N')

        for n in ns:
            protons = [(b[self.ccbAtomId1] if b[self.ccbAtomId1] != n else b[self.ccbAtomId2])
                       for b in self.lastBonds
                       if (b[self.ccbAtomId1] == n and b[self.ccbAtomId2][0] in protonBeginCode)
                       or (b[self.ccbAtomId2] == n and b[self.ccbAtomId1][0] in protonBeginCode)]
            if len(protons) != 2:
                continue
            atmList.append(protons[0])

        return atmList

    def getNonRepAminoProtons(self, compId):
        """ Return non-representative protons in amino group of a given comp_id.
        """

        if compId != self.lastCompId and not self.updateChemCompDict(compId):
            return []

        atmList = []

        ns = (a[self.ccaAtomId] for a in self.lastAtomList if a[self.ccaTypeSymbol] == 'N')

        for n in ns:
            protons = [(b[self.ccbAtomId1] if b[self.ccbAtomId1] != n else b[self.ccbAtomId2])
                       for b in self.lastBonds
                       if (b[self.ccbAtomId1] == n and b[self.ccbAtomId2][0] in protonBeginCode)
                       or (b[self.ccbAtomId2] == n and b[self.ccbAtomId1][0] in protonBeginCode)]
            if len(protons) != 2:
                continue
            atmList.extend(protons[1:])

        return atmList

    def getImideProtons(self, compId):
        """ Return imide protons of a given comp_id.
        """

        if compId != self.lastCompId and not self.updateChemCompDict(compId):
            return []

        atmList = []

        ns = (a[self.ccaAtomId] for a in self.lastAtomList if a[self.ccaTypeSymbol] == 'N')

        for n in ns:
            protons = [(b[self.ccbAtomId1] if b[self.ccbAtomId1] != n else b[self.ccbAtomId2])
                       for b in self.lastBonds
                       if (b[self.ccbAtomId1] == n and b[self.ccbAtomId2][0] in protonBeginCode)
                       or (b[self.ccbAtomId2] == n and b[self.ccbAtomId1][0] in protonBeginCode)]
            if len(protons) != 1:
                continue
            acyl_c = [(b[self.ccbAtomId1] if b[self.ccbAtomId1] != n else b[self.ccbAtomId2])
                      for b in self.lastBonds
                      if (b[self.ccbAtomId1] == n and b[self.ccbAtomId2][0] == 'C')
                      or (b[self.ccbAtomId2] == n and b[self.ccbAtomId1][0] == 'C')]
            if len(acyl_c) != 2:
                continue
            acyl_o = [(b[self.ccbAtomId1] if b[self.ccbAtomId1] not in acyl_c else b[self.ccbAtomId2])
                      for b in self.lastBonds if b[self.ccbValueOrder] == 'DOUB'
                      and ((b[self.ccbAtomId1] in acyl_c and b[self.ccbAtomId2][0] == 'O')
                           or (b[self.ccbAtomId2] in acyl_c and b[self.ccbAtomId1][0] == 'O'))]
            if len(acyl_o) != 2:
                continue
            atmList.append(protons[0])

        return atmList

    def getBondedAtoms(self, compId, atomId, exclProton=False, onlyProton=False):
        """ Return bonded atoms to a given atom.
        """

        if compId != self.lastCompId and not self.updateChemCompDict(compId):
            return []

        bondedAtoms = [(b[self.ccbAtomId1] if b[self.ccbAtomId1] != atomId else b[self.ccbAtomId2])
                       for b in self.lastBonds if atomId in (b[self.ccbAtomId1], b[self.ccbAtomId2])]

        if not exclProton and not onlyProton:
            return bondedAtoms

        allProtons = [a[self.ccaAtomId] for a in self.lastAtomList if a[self.ccaTypeSymbol] == 'H']

        return [a for a in bondedAtoms if (exclProton and a not in allProtons) or (onlyProton and a in allProtons)]

    def getProtonsInSameGroup(self, compId, atomId, exclSelf=False):
        """ Return protons in the same group of a given comp_id and atom_id.
        """

        if compId != self.lastCompId and not self.updateChemCompDict(compId):
            return []

        bondedTo = self.getBondedAtoms(compId, atomId)

        if len(bondedTo) == 0:
            return []

        attached = self.getBondedAtoms(compId, bondedTo[0], onlyProton=True)

        return [p for p in attached if (exclSelf and p != atomId) or not exclSelf]

    def getAtomsBasedOnGreekLetterSystem(self, compId, atomId):
        """ Return atoms match with greek letter system of a given comp_id.
        """

        if len(atomId) < 2:
            return {}

        elem = atomId[0]

        if elem not in ('H', 'C', 'N', 'O', 'S', 'P'):
            return {}

        greekLetter = atomId[1]

        if greekLetter not in ('A', 'B', 'G', 'D', 'E', 'Z', 'H'):
            return {}

        if compId != self.lastCompId and not self.updateChemCompDict(compId):
            return {}

        touched = ['N', 'C', 'O', 'OXT']
        parents = ['CA']

        if not any(a[self.ccaAtomId] == 'CA' for a in self.lastAtomList):
            if not any(a[self.ccaAtomId] == 'C' for a in self.lastAtomList):  # ACA:QA -> H21, H22 (5nwu)
                return {}
            bonded = [(b[self.ccbAtomId1] if b[self.ccbAtomId1] != 'C' else b[self.ccbAtomId2])
                      for b in self.lastBonds if 'C' in (b[self.ccbAtomId1], b[self.ccbAtomId2])]
            if len(bonded) == 0:
                return {}
            parents = [b for b in bonded if b not in touched]
            if len(parents) == 0:
                return {}

        for letter in ['A', 'B', 'G', 'D', 'E', 'Z', 'H']:

            if greekLetter == letter:
                atoms = parents
                for p in parents:
                    atoms.extend(self.getBondedAtoms(compId, p, onlyProton=True))
                return {a for a in atoms if a[0] == elem}

            touched.extend(parents)

            _parents = []
            for p in parents:
                _parents.extend([a for a in self.getBondedAtoms(compId, p, exclProton=True) if a not in touched])

            if len(_parents) == 0:
                return {}

            parents = list(set(_parents))

        return {}

    def hasBond(self, compId, atomId1, atomId2):
        """ Return whether given two atoms are connected by a covalent bond.
        """
        return atomId2 in self.getBondedAtoms(compId, atomId1)

    def peptideLike(self, compId=None):
        """ Return whether a given comp_id is peptide-like component.
        """

        if compId is not None:
            if not self.updateChemCompDict(compId):
                return False

        ctype = self.lastChemCompDict['_chem_comp.type'].upper()

        if 'PEPTIDE' in ctype:
            return True

        if 'DNA' in ctype or 'RNA' in ctype or 'SACCHARIDE' in ctype:
            return False

        peptide_like = len([a for a in self.lastAtomList
                            if a[self.ccaAtomId] in ("C", "CA", "H", "HA", "HA2", "HA3", "N", "O")])

        nucleotide_like = len([a for a in self.lastAtomList
                               if a[self.ccaAtomId] in ("C1'", "C2'", "C3'", "C4'", "C5'",
                                                        "H1'", "H2'", "H2''", "H3'", "H4'", "H5'", "H5''", "HO2'",
                                                        "P", "OP1", "OP2", "O5'", "O3'")])

        carbohydrate_like = len([a for a in self.lastAtomList
                                 if a[self.ccaAtomId] in ("C1", "C2", "C3", "C4", "C5", "C6",
                                                          "H1", "H2", "H3", "H4", "H5", "H61", "H62",
                                                          "O1", "O4", "O6")])

        return peptide_like > nucleotide_like and peptide_like > carbohydrate_like

    def getTypeOfCompId(self, compId=None):
        """ Return type of a given comp_id.
            @return: array of bool: peptide, nucleotide, carbohydrate
        """

        if compId is not None:
            if not self.updateChemCompDict(compId):
                return False, False, False

        results = [False] * 3

        ctype = self.lastChemCompDict['_chem_comp.type'].upper()

        if 'PEPTIDE' in ctype:
            results[0] = True
            return results

        if 'DNA' in ctype or 'RNA' in ctype:
            results[1] = True
            return results

        if 'SACCHARIDE' in ctype:
            results[2] = True
            return results

        peptide_like = len([a for a in self.lastAtomList
                            if a[self.ccaAtomId] in ("C", "CA", "H", "HA", "HA2", "HA3", "N", "O")])

        nucleotide_like = len([a for a in self.lastAtomList
                               if a[self.ccaAtomId] in ("C1'", "C2'", "C3'", "C4'", "C5'",
                                                        "H1'", "H2'", "H2''", "H3'", "H4'", "H5'", "H5''", "HO2'",
                                                        "P", "OP1", "OP2", "O5'", "O3'")])

        carbohydrate_like = len([a for a in self.lastAtomList
                                 if a[self.ccaAtomId] in ("C1", "C2", "C3", "C4", "C5", "C6",
                                                          "H1", "H2", "H3", "H4", "H5", "H61", "H62",
                                                          "O1", "O4", "O6")])

        results[0] = peptide_like > nucleotide_like and peptide_like > carbohydrate_like
        results[1] = nucleotide_like > peptide_like and nucleotide_like > carbohydrate_like
        results[2] = carbohydrate_like > peptide_like and carbohydrate_like > nucleotide_like

        return results

    def getEffectiveFormulaWeight(self, compId):
        """ Return effective formula weight of a given comp_id.
        """

        if not self.updateChemCompDict(compId):
            return 0.0

        if '_chem_comp.formula_weight' not in self.lastChemCompDict:
            return 0.0

        try:

            fw = float(self.lastChemCompDict['_chem_comp.formula_weight'])

        except ValueError:
            return 0.0

        peptide_like = self.peptideLike()

        leavingTypeSymbols = [a[self.ccaTypeSymbol] for a in self.lastAtomList
                              if not (a[self.ccaLeavingAtomFlag] != 'Y'
                                      or (peptide_like
                                          and a[self.ccaNTerminalAtomFlag] == 'N'
                                          and a[self.ccaCTerminalAtomFlag] == 'N'))]

        try:

            for symbol in leavingTypeSymbols:
                fw -= ELEMENT_WEIGHTS[NAMES_ELEMENT[symbol.title()]]

        except KeyError:
            return 0.0

        return fw

    def write_std_dict_as_pickle(self):
        """ Write dictionary for standard residues as pickle file.
        """

        for compId in monDict3:
            self.updateChemCompDict(compId)

        def write_dict_as_pickle(obj, file_name):
            """ Write dictionary as pickle file.
            """

            if isinstance(obj, dict):

                with open(file_name, 'wb') as ofh:
                    pickle.dump(obj, ofh)

        write_dict_as_pickle(self.__cachedDict, self.__cacheFile)



"""
File:    mmCIFUtil.py
Author:  Zukang Feng
Update:  21-August-2012
Version: 001  Initial version
# Update:
# 07-Apr-2020  M. Yokochi - Re-write Zukang's version to being aware of multiple data blocks
# 30-May-2024  M. Yokochi - Resolve duplication of data block/saveframe name (DAOTHER-9437)
##
"""

__author__ = "Zukang Feng"
__email__ = "zfeng@rcsb.rutgers.edu"
__version__ = "V0.001"

import sys
import copy

from mmcif.api.DataCategory import DataCategory
from mmcif.api.PdbxContainers import DataContainer
from mmcif.io.PdbxReader import PdbxReader
from mmcif.io.PdbxWriter import PdbxWriter


def get_ext_block_name(name, ext=1):
    """ Return unique block name avoiding duplication
    """
    return name if ext == 1 else f'{name}_{ext}'


class mmCIFUtil:
    """ Using pdbx mmCIF utility to parse mmCIF file
    """

    def __init__(self, verbose=False, log=sys.stderr, filePath=None):  # pylint: disable=unused-argument
        # self.__verbose = verbose
        self.__lfh = log
        self.__filePath = filePath
        self.__dataList = []
        self.__dataMap = {}
        self.__blockNameList = []
        self.__read()

    def __read(self):
        if not self.__filePath:
            return
        #
        try:
            with open(self.__filePath, 'r', encoding='utf-8') as ifh:
                pRd = PdbxReader(ifh)
                pRd.read(self.__dataList)

            if self.__dataList:
                is_star = all(container.getType() == 'data' for container in self.__dataList)
                idx = 0
                for container in self.__dataList:
                    if is_star or (not is_star and container.getType() != 'data'):
                        blockName = container.getName()
                        self.__blockNameList.append(blockName)

                        if blockName not in self.__dataMap:
                            self.__dataMap[blockName] = idx
                        else:
                            ext = 2
                            while True:
                                extBlockName = get_ext_block_name(blockName, ext)
                                if extBlockName not in self.__dataMap:
                                    self.__dataMap[extBlockName] = idx
                                    break
                                ext += 1
                    idx += 1
                #
            #
        except Exception as e:
            self.__lfh.write(f"Read {self.__filePath} failed {str(e)}.\n")

    def GetBlockIDList(self):
        """ Return list of block ID
        """
        return self.__blockNameList

    def GetValueAndItemByBlock(self, blockName, catName, ext=1):
        """ Get category values and item names
        """
        dList = []
        iList = []
        if blockName not in self.__dataMap:
            return dList, iList
        #
        catObj = self.__dataList[self.__dataMap[get_ext_block_name(blockName, ext)]].getObj(catName)
        if not catObj:
            return dList, iList
        #
        iList = catObj.getAttributeList()
        rowList = catObj.getRowList()
        for row in rowList:
            try:
                tD = {}
                for idxIt, itName in enumerate(iList):
                    if row[idxIt] != "?" and row[idxIt] != ".":
                        tD[itName] = row[idxIt]
                #
                if tD:
                    dList.append(tD)
                #
            except IndexError:
                pass
        #
        return dList, iList

    def GetValue(self, blockName, catName, ext=1):
        """ Get category values in a given Data Block and Category
            The results are stored in a list of dictionaries with item name as key
        """
        return self.GetValueAndItemByBlock(blockName, catName, ext)[0]

    def GetSingleValue(self, blockName, catName, itemName, ext):
        """ Get the first value of a given Data Block, Category, Item
        """
        text = ""
        dlist = self.GetValue(blockName, catName, ext)
        if dlist:
            if itemName in dlist[0]:
                text = dlist[0][itemName]
        return text

    def UpdateSingleRowValue(self, blockName, catName, itemName, rowIndex, value, ext=1):
        """ Update value in single row
        """
        if blockName not in self.__dataMap:
            return

        catObj = self.__dataList[self.__dataMap[get_ext_block_name(blockName, ext)]].getObj(catName)

        if not catObj:
            return
        #
        catObj.setValue(value, itemName, rowIndex)

    def UpdateMultipleRowsValue(self, blockName, catName, itemName, value, ext=1):
        """ Update value in multiple rows
        """
        if blockName not in self.__dataMap:
            return

        catObj = self.__dataList[self.__dataMap[get_ext_block_name(blockName, ext)]].getObj(catName)

        if not catObj:
            return
        #
        rowNo = catObj.getRowCount()
        for rowIndex in range(0, rowNo):
            catObj.setValue(value, itemName, rowIndex)

    def AddBlock(self, blockName, ext=1):
        """ Add Data Block
        """
        container = DataContainer(blockName)
        self.__dataMap[get_ext_block_name(blockName, ext)] = len(self.__dataList)
        self.__dataList.append(container)

    def AddCategory(self, blockName, catName, items, ext=1):
        """ Add Category in a given Data Block
        """
        if blockName not in self.__dataMap:
            return

        category = DataCategory(catName)
        for item in items:
            category.appendAttribute(item)
        #
        self.__dataList[self.__dataMap[get_ext_block_name(blockName, ext)]].append(category)

    def RemoveCategory(self, blockName, catName, ext=1):
        """ Remove Category in a given Data Block
        """
        if blockName not in self.__dataMap:
            return

        idx = self.__dataMap[get_ext_block_name(blockName, ext)]

        catObj = self.__dataList[idx].getObj(catName)

        if catObj is None:
            return

        self.__dataList[idx].remove(catName)

    def MoveCategoryToTop(self, blockName, catName, ext=1):
        """ Move Category to top in a given Data Block
        """
        if blockName not in self.__dataMap:
            return

        idx = self.__dataMap[get_ext_block_name(blockName, ext)]

        catObj = self.__dataList[idx].getObj(catName)

        if catObj is None:
            return

        _catNameList = copy.copy(self.__dataList[idx].getObjNameList())
        _catNameList.remove(catName)

        _catObjList = [copy.copy(catObj)]
        _catObjList.extend([copy.copy(self.__dataList[idx].getObj(_catName)) for _catName in _catNameList])

        for _catName in _catNameList:
            self.__dataList[idx].remove(_catName)

        for _catObj in _catObjList:
            self.__dataList[idx].append(_catObj)

    def InsertData(self, blockName, catName, dataList, ext=1):
        """ Insert data in a given Data Block and Category
        """
        if blockName not in self.__dataMap:
            return

        catObj = self.__dataList[self.__dataMap[get_ext_block_name(blockName, ext)]].getObj(catName)

        if catObj is None:
            return
        #
        for data in dataList:
            catObj.append(data)

    def ExtendCategory(self, blockName, catName, items, dataList, col=-1, ext=1):
        """ Extend existing Category in a given Data Block
        """
        if blockName not in self.__dataMap:
            return

        catObj = self.__dataList[self.__dataMap[get_ext_block_name(blockName, ext)]].getObj(catName)

        if catObj is None:
            return

        append_items = col < 0 or col >= catObj.getAttributeCount()

        data_len = len(dataList)
        empty_row = ['.'] * len(items)

        if append_items:

            for item in items:
                catObj.appendAttribute(item)

            rowList = catObj.getRowList()

            for idx, row in enumerate(rowList):

                if idx < data_len:
                    row.extend(dataList[idx])
                else:
                    row.extend(empty_row)

        else:

            attrNameList = catObj.getAttributeList()

            _attrNameList = attrNameList[0:col] + items + attrNameList[col:]

            rowList = catObj.getRowList()

            _rowList = []

            for idx, row in enumerate(rowList):
                _row = row[0:col]

                if idx < data_len:
                    _row.extend(dataList[idx])
                else:
                    _row.extend(empty_row)

                _row.extend(row[col:])

                _rowList.append(_row)

            catObj.setAttributeNameList(_attrNameList)
            catObj.setRowList(_rowList)

    def CopyValueInRow(self, blockName, catName, srcItems, dstItems, ext=1):
        """ Copy value from source items to destination items
        """
        if srcItems is None or dstItems is None or len(srcItems) != len(dstItems):
            return

        if blockName not in self.__dataMap:
            return

        catObj = self.__dataList[self.__dataMap[get_ext_block_name(blockName, ext)]].getObj(catName)

        if catObj is None:
            return

        attrs = catObj.getAttributeList()

        for item in srcItems:
            if item not in attrs:
                return

        for item in dstItems:
            if item not in attrs:
                return

        src_cols = [attrs.index(item) for item in srcItems]
        dst_cols = [attrs.index(item) for item in dstItems]

        for row in catObj.getRowList():
            for j, src_col in enumerate(src_cols):
                row[dst_cols[j]] = row[src_col]

    def WriteCif(self, outputFilePath=None):
        """ Write out cif file
        """
        if not outputFilePath:
            return
        #
        with open(outputFilePath, 'w', encoding='utf-8') as ofh:
            pdbxW = PdbxWriter(ofh)
            pdbxW.write(self.__dataList)

    def GetCategories(self):
        """ Get all Categories in all Data Blocks
        """
        data = {}
        for container in self.__dataList:
            blockName = container.getName()
            data[blockName] = container.getObjNameList()
        #
        return data

    def GetAttributes(self, blockName, catName, ext=1):
        """ Get item name in Data Block and Category
        """
        if blockName not in self.__dataMap:
            return []

        catObj = self.__dataList[self.__dataMap[get_ext_block_name(blockName, ext)]].getObj(catName)

        if not catObj:
            return []
        #
        return catObj.getAttributeList()

    def GetDictList(self, blockName, catName, ext=1):
        """ Get a list of dictionaries of a given Data Block and Category
        """
        dList, iList = self.GetValueAndItemByBlock(blockName, catName, ext)
        data = [[x.get(y) for y in iList] for x in dList]
        #
        return {catName: {"Items": iList, "Values": data}}

    def GetDataBlock(self, blockName, ext=1):
        """ Get a list of dictionaries of a given Data Block
        """
        data = {}
        categories = self.GetCategories()
        if blockName not in categories:
            return data
        #
        for catName in categories[blockName]:
            data.update(self.GetDictList(blockName, catName, ext))
        #
        return data



##
# BiosymMRReader.py
#
# Update:
##
""" A collection of classes for parsing BIOSYM MR files.
"""
import sys
import os

from antlr4 import InputStream, CommonTokenStream, ParseTreeWalker, PredictionMode

try:
    from wwpdb.utils.nmr.mr.LexerErrorListener import LexerErrorListener
    from wwpdb.utils.nmr.mr.ParserErrorListener import ParserErrorListener
    from wwpdb.utils.nmr.mr.BiosymMRLexer import BiosymMRLexer
    from wwpdb.utils.nmr.mr.BiosymMRParser import BiosymMRParser
    from wwpdb.utils.nmr.mr.BiosymMRParserListener import BiosymMRParserListener
    from wwpdb.utils.nmr.mr.ParserListenerUtil import (coordAssemblyChecker,
                                                       MAX_ERROR_REPORT,
                                                       REPRESENTATIVE_MODEL_ID,
                                                       REPRESENTATIVE_ALT_ID)
    from wwpdb.utils.nmr.io.CifReader import CifReader
    from wwpdb.utils.nmr.ChemCompUtil import ChemCompUtil
    from wwpdb.utils.nmr.BMRBChemShiftStat import BMRBChemShiftStat
    from wwpdb.utils.nmr.NEFTranslator.NEFTranslator import NEFTranslator
except ImportError:
    from nmr.mr.LexerErrorListener import LexerErrorListener
    from nmr.mr.ParserErrorListener import ParserErrorListener
    from nmr.mr.BiosymMRLexer import BiosymMRLexer
    from nmr.mr.BiosymMRParser import BiosymMRParser
    from nmr.mr.BiosymMRParserListener import BiosymMRParserListener
    from nmr.mr.ParserListenerUtil import (coordAssemblyChecker,
                                           MAX_ERROR_REPORT,
                                           REPRESENTATIVE_MODEL_ID,
                                           REPRESENTATIVE_ALT_ID)
    from nmr.io.CifReader import CifReader
    from nmr.ChemCompUtil import ChemCompUtil
    from nmr.BMRBChemShiftStat import BMRBChemShiftStat
    from nmr.NEFTranslator.NEFTranslator import NEFTranslator


class BiosymMRReader:
    """ Accessor methods for parsing BIOSYM MR files.
    """

    def __init__(self, verbose=True, log=sys.stdout,
                 representativeModelId=REPRESENTATIVE_MODEL_ID,
                 representativeAltId=REPRESENTATIVE_ALT_ID,
                 mrAtomNameMapping=None,
                 cR=None, caC=None, ccU=None, csStat=None, nefT=None,
                 reasons=None):
        self.__verbose = verbose
        self.__lfh = log
        self.__debug = False

        self.__maxLexerErrorReport = MAX_ERROR_REPORT
        self.__maxParserErrorReport = MAX_ERROR_REPORT

        self.__representativeModelId = representativeModelId
        self.__representativeAltId = representativeAltId
        self.__mrAtomNameMapping = mrAtomNameMapping

        if cR is not None and caC is None:
            caC = coordAssemblyChecker(verbose, log, representativeModelId, representativeAltId,
                                       cR, None, None, fullCheck=False)

        self.__cR = cR
        self.__caC = caC

        # CCD accessing utility
        self.__ccU = ChemCompUtil(verbose, log) if ccU is None else ccU

        # BMRB chemical shift statistics
        self.__csStat = BMRBChemShiftStat(verbose, log, self.__ccU) if csStat is None else csStat

        # NEFTranslator
        self.__nefT = NEFTranslator(verbose, log, self.__ccU, self.__csStat) if nefT is None else nefT
        if nefT is None:
            self.__nefT.set_remediation_mode(True)

        # reasons for re-parsing request from the previous trial
        self.__reasons = reasons

    def setDebugMode(self, debug):
        self.__debug = debug

    def setLexerMaxErrorReport(self, maxErrReport):
        self.__maxLexerErrorReport = maxErrReport

    def setParserMaxErrorReport(self, maxErrReport):
        self.__maxParserErrorReport = maxErrReport

    def parse(self, mrFilePath, cifFilePath=None, isFilePath=True,
              createSfDict=False, originalFileName=None, listIdCounter=None, entryId=None):
        """ Parse BIOSYM MR file.
            @return: BiosymMRParserListener for success or None otherwise, ParserErrorListener, LexerErrorListener.
        """

        ifh = None

        try:

            if isFilePath:
                mrString = None

                if not os.access(mrFilePath, os.R_OK):
                    if self.__verbose:
                        self.__lfh.write(f"BiosymMRReader.parse() {mrFilePath} is not accessible.\n")
                    return None, None, None

                ifh = open(mrFilePath, 'r')  # pylint: disable=consider-using-with
                input = InputStream(ifh.read())

            else:
                mrFilePath, mrString = None, mrFilePath

                if mrString is None or len(mrString) == 0:
                    if self.__verbose:
                        self.__lfh.write("BiosymMRReader.parse() Empty string.\n")
                    return None, None, None

                input = InputStream(mrString)

            if cifFilePath is not None:
                if not os.access(cifFilePath, os.R_OK):
                    if self.__verbose:
                        self.__lfh.write(f"BiosymMRReader.parse() {cifFilePath} is not accessible.\n")
                    return None, None, None

                if self.__cR is None:
                    self.__cR = CifReader(self.__verbose, self.__lfh)
                    if not self.__cR.parse(cifFilePath):
                        return None, None, None

            lexer = BiosymMRLexer(input)
            lexer.removeErrorListeners()

            lexer_error_listener = LexerErrorListener(mrFilePath, maxErrorReport=self.__maxLexerErrorReport)
            lexer.addErrorListener(lexer_error_listener)

            messageList = lexer_error_listener.getMessageList()

            if messageList is not None and self.__verbose:
                for description in messageList:
                    self.__lfh.write(f"[Syntax error] line {description['line_number']}:{description['column_position']} {description['message']}\n")
                    if 'input' in description:
                        self.__lfh.write(f"{description['input']}\n")
                        self.__lfh.write(f"{description['marker']}\n")

            stream = CommonTokenStream(lexer)
            parser = BiosymMRParser(stream)
            # try with simpler/faster SLL prediction mode
            parser._interp.predictionMode = PredictionMode.SLL  # pylint: disable=protected-access
            parser.removeErrorListeners()
            parser_error_listener = ParserErrorListener(mrFilePath, maxErrorReport=self.__maxParserErrorReport)
            parser.addErrorListener(parser_error_listener)
            tree = parser.biosym_mr()

            walker = ParseTreeWalker()
            listener = BiosymMRParserListener(self.__verbose, self.__lfh,
                                              self.__representativeModelId,
                                              self.__representativeAltId,
                                              self.__mrAtomNameMapping,
                                              self.__cR, self.__caC,
                                              self.__ccU, self.__csStat, self.__nefT,
                                              self.__reasons)
            listener.setDebugMode(self.__debug)
            listener.createSfDict(createSfDict)
            if createSfDict:
                if originalFileName is not None:
                    listener.setOriginaFileName(originalFileName)
                if listIdCounter is not None:
                    listener.setListIdCounter(listIdCounter)
                if entryId is not None:
                    listener.setEntryId(entryId)
            walker.walk(listener, tree)

            messageList = parser_error_listener.getMessageList()

            if messageList is not None and self.__verbose:
                for description in messageList:
                    self.__lfh.write(f"[Syntax error] line {description['line_number']}:{description['column_position']} {description['message']}\n")
                    if 'input' in description:
                        self.__lfh.write(f"{description['input']}\n")
                        self.__lfh.write(f"{description['marker']}\n")

            if self.__verbose:
                if listener.warningMessage is not None and len(listener.warningMessage) > 0:
                    print('\n'.join(listener.warningMessage))
                if isFilePath:
                    print(listener.getContentSubtype())

            return listener, parser_error_listener, lexer_error_listener

        except IOError as e:
            if self.__verbose:
                self.__lfh.write(f"+BiosymMRReader.parse() ++ Error - {str(e)}\n")
            return None, None, None
            # pylint: disable=unreachable
            """ debug code
        except Exception as e:
            if self.__verbose and isFilePath:
                self.__lfh.write(f"+BiosymMRReader.parse() ++ Error - {mrFilePath!r} - {str(e)}\n")
            return None, None, None
            """
        finally:
            if isFilePath and ifh is not None:
                ifh.close()


if __name__ == "__main__":
    reader = BiosymMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2vb5/2vb5-corrected-div_dst-div_src.mr',
                 '../../tests-nmr/mock-data-remediation/2vb5/2vb5.cif')

    reader = BiosymMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2lyw/2lyw-trimmed.mr',
                 '../../tests-nmr/mock-data-remediation/2lyw/2lyw.cif')

    reader = BiosymMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/6var/HBVe_SAXS.tab',
                 '../../tests-nmr/mock-data-remediation/6var/6var.cif')

    reader = BiosymMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2mug/2mug-trimmed.mr',
                 '../../tests-nmr/mock-data-remediation/2mug/2mug.cif')



##
# File: CifToNmrStar.py
# Date: 19-Jul-2021
#
# Updates:
# 13-Oct-2021  M. Yokochi - code revision according to PEP8 using Pylint (DAOTHER-7389, issue #5)
# 20-Apr-2022  M. Yokochi - enable to fix broken data block order of CIF formatted NMR-STAR using NMR-STAR schema (DAOTHER-7407, NMR restraint remediation)
# 28-Jul-2022  M. Yokochi - enable to fix format issue of CIF formatted NMR-STAR (You cannot have two loops with the same category in one saveframe. Category: '_Audit')
# 27-Sep-2022  M. Yokochi - auto fill list ID and entry ID (NMR restraint remediation)
# 13-Jun-2023  M. Yokochi - sort loops in a saveframe based on schema
# 30-May-2024  M. Yokochi - resolve duplication of data block/saveframe name (DAOTHER-9437)
# 25-Jun-2024  M. Yokochi - strip white spaces in a datablock name derived from the model file (DAOTHER-9511)
##
""" Wrapper class for CIF to NMR-STAR converter.
    @author: Masashi Yokochi
"""
import sys
import os
import re
import pynmrstar
import pickle
import logging
import hashlib
import collections

from packaging import version
from operator import itemgetter

try:
    from wwpdb.utils.nmr.io.mmCIFUtil import mmCIFUtil
    from wwpdb.utils.nmr.AlignUtil import (emptyValue, trueValue)
except ImportError:
    from nmr.io.mmCIFUtil import mmCIFUtil
    from nmr.AlignUtil import (emptyValue, trueValue)


__pynmrstar_v3_3_1__ = version.parse(pynmrstar.__version__) >= version.parse("3.3.1")
__pynmrstar_v3_2__ = version.parse(pynmrstar.__version__) >= version.parse("3.2.0")
__pynmrstar_v3_1__ = version.parse(pynmrstar.__version__) >= version.parse("3.1.0")
__pynmrstar_v3__ = version.parse(pynmrstar.__version__) >= version.parse("3.0.0")

if __pynmrstar_v3_3_1__:
    logger = logging.getLogger('pynmrstar')
    logger.setLevel(logging.ERROR)
else:
    logging.getLogger().setLevel(logging.ERROR)  # set level for pynmrstar


class CifToNmrStar:
    """ Simple CIF to NMR-STAR converter.
    """

    def __init__(self, log=sys.stderr):
        self.__lfh = log

        # directory
        self.schema_dir = os.path.dirname(__file__) + '/nmr-star_schema/'

        def load_schema_from_pickle(file_name):
            """ Load NMR-STAR schema from pickle file.
            """

            if os.path.exists(file_name):

                with open(file_name, 'rb') as ifh:
                    return pickle.load(ifh)

            return None

        # NMR-STAR schema
        self.schema = load_schema_from_pickle(self.schema_dir + 'schema.pkl')

        # NMR-STAR schema order
        # self.schema_order = load_schema_from_pickle(self.schema_dir + 'schema_order.pkl')

        # NMR-STAR category order
        self.category_order = load_schema_from_pickle(self.schema_dir + 'category_order.pkl')

        if self.schema is None:
            schema = pynmrstar.Schema()  # network latency occurs if pickle resource files are not available
            self.schema = schema.schema
            self.category_order = schema.category_order

        # NEF category order
        self.category_order_nef = ['_nef_nmr_meta_data',
                                   '_nef_molecular_system',
                                   '_nef_chemical_shift_list',
                                   '_nef_distance_restraint_list',
                                   '_nef_dihedral_restraint_list',
                                   '_nef_rdc_restraint_list',
                                   '_nef_nmr_spectrum',
                                   '_nef_peak_restraint_links']

    def write_schema_as_pickles(self):
        """ Retrieve NMR-STAR schema from pynmrstar.Schema, then write schema objects as each pickle file.
        """

        def write_schema_as_pickle(obj, file_name):
            """ Write NMR-STAR schema as pickle file.
            """

            with open(file_name, 'wb') as ofh:
                pickle.dump(obj, ofh)

        schema = pynmrstar.Schema()  # retrieve the latest schema via interet access

        # print(schema.headers)
        with open(self.schema_dir + 'headers.txt', 'w') as ofh:
            for header in schema.headers:
                ofh.write(header + '\n')
        self.__lfh.write('headers.txt: Done.\n')

        # print(schema.schema)
        write_schema_as_pickle(schema.schema, self.schema_dir + 'schema.pkl')
        self.__lfh.write('schema.pkl: Done.\n')

        # print(schema.schema_order)
        write_schema_as_pickle(schema.schema_order, self.schema_dir + 'schema_order.pkl')
        self.__lfh.write('schema_order.pkl: Done.\n')

        # print(schema.category_order)
        write_schema_as_pickle(schema.category_order, self.schema_dir + 'category_order.pkl')
        self.__lfh.write('category_order.pkl: Done.\n')

        # print(schema.data_types)
        write_schema_as_pickle(schema.data_types, self.schema_dir + 'data_types.pkl')
        self.__lfh.write('data_types.pkl: Done.\n')

        with open(self.schema_dir + 'version.txt', 'w') as ofh:
            ofh.write(schema.version)
        self.__lfh.write(f"version: {schema.version}\n")

    def convert(self, cifPath=None, strPath=None, datablockName=None, maxRepeat=1):
        """ Convert CIF formatted NMR data file to normalized NMR-STAR file.
        """

        if cifPath is None or strPath is None:
            return False

        try:

            cifObj = mmCIFUtil(filePath=cifPath)

            block_name_list = cifObj.GetBlockIDList()

            if len(block_name_list) == 0:  # single loop
                return False

            dup_block_name_list = []
            if len(block_name_list) > 1:
                c = collections.Counter(block_name_list).most_common()
                dup_block_name_list = [k for k, v in c if v > 1]

            block_name_pat_w_list_0 = [block_name[:-1] for block_name in block_name_list if block_name.endswith('list_0')]

            block_name_counter = {n: 0 for n in block_name_list}

            entry_id = None

            strData = pynmrstar.Entry.from_scratch(datablockName if datablockName is not None else os.path.basename(cifPath))

            # check category order in CIF
            category_order = []
            previous_order = -1

            sf_category_counter = {}

            for block_name in block_name_list:
                block_name_counter[block_name] += 1
                ext = block_name_counter[block_name]

                dict_list = cifObj.GetDataBlock(block_name, ext)

                ordered_block = True
                sf_category = ''

                for category, itVals in dict_list.items():
                    try:
                        current_order = self.category_order.index('_' + category)
                    except ValueError:
                        continue
                    # print(f"{block_name} {category} {current_order}")
                    item = {'block_name': block_name, 'category': category,
                            'category_order': current_order,
                            'block_name_ext': ext}
                    if current_order < previous_order:
                        ordered_block = False
                        for item2 in category_order:
                            if item2['block_name'] == block_name:
                                item2['ordered'] = False
                    item['ordered'] = ordered_block

                    item['sf_category_flag'] = False
                    for _item, _value in zip(itVals['Items'], itVals['Values'][0]):
                        tag = '_' + category + '.' + _item
                        tag = tag.lower()
                        if tag in self.schema:
                            sdict = {k: v for k, v in self.schema[tag].items() if v not in emptyValue}
                            if 'sf_category' not in item:
                                item['sf_category'] = sdict['SFCategory']
                            if 'super_category' not in item:
                                item['super_category'] = sdict['ADIT super category']
                            if 'Sf category flag' in sdict and sdict['Sf category flag'].lower() in trueValue:
                                item['sf_category_flag'] = True
                            if _item == 'Entry_ID' and entry_id in emptyValue:
                                entry_id = _value

                    if 'Sf_category' not in itVals['Items']:
                        _tag = '_' + category.lower() + '.sf_category'
                        if _tag in self.schema:
                            sdict = {k: v for k, v in self.schema[_tag].items() if v not in emptyValue}
                            if 'sf_category' not in item:
                                item['sf_category'] = sdict['SFCategory']
                            if 'super_category' not in item:
                                item['super_category'] = sdict['ADIT super category']
                            if 'Sf category flag' in sdict and sdict['Sf category flag'].lower() in trueValue:
                                item['sf_category_flag'] = True

                    if len(sf_category) == 0:
                        sf_category = item['sf_category']
                    elif sf_category != item['sf_category']:
                        for item2 in category_order:
                            if item2['block_name'] == block_name:
                                item2['ordered'] = False
                        item['ordered'] = False

                    if item not in category_order:
                        category_order.append(item)

                    if item['sf_category_flag']:
                        previous_order = current_order

            if entry_id in emptyValue:
                entry_id = block_name_list[0].strip().replace(' ', '_')  # DAOTHER-9511: replace white space in a datablock name to underscore

            _entry_id = entry_id.upper()

            if datablockName is None:
                strData.entry_id = f'cs_{entry_id.lower()}'

            # reorder
            category_order.sort(key=itemgetter('category_order'))

            # correct block name in case
            reserved_block_names = []
            prev_sf_category = ''
            prev_sf_block_name = ''
            for item in category_order:
                block_name = item['block_name']
                split_block_name = block_name.split('_')

                if block_name in dup_block_name_list:
                    if split_block_name[-1].isdigit():
                        split_block_name.pop()
                    item['new_block_name'] = '_'.join(split_block_name) + '_' + str(item['block_name_ext'])

                elif any(block_name.startswith(pat) for pat in block_name_pat_w_list_0):  # STARCh output
                    if split_block_name[-1].isdigit():
                        list_id = int(split_block_name[-1]) + 1
                        split_block_name.pop()
                        item['new_block_name'] = '_'.join(split_block_name) + '_' + str(list_id)

                if item['ordered']:
                    continue

                if item['sf_category_flag']:
                    if item['sf_category'] == item['super_category']:
                        block_name = item['sf_category']
                    else:
                        list_id = 1
                        while True:
                            block_name = item['sf_category'] + '_' + str(list_id)
                            if block_name not in reserved_block_names:
                                break
                            list_id += 1
                    prev_sf_category = item['sf_category']
                    prev_sf_block_name = block_name
                else:
                    if item['sf_category'] == prev_sf_category:
                        block_name = prev_sf_block_name
                    else:
                        item['missing_sf_category'] = True
                        if item['sf_category'] == item['super_category']:
                            block_name = item['sf_category']
                        else:
                            list_id = 1
                            while True:
                                block_name = item['sf_category'] + '_' + str(list_id)
                                if block_name not in reserved_block_names:
                                    break
                                list_id += 1
                        prev_sf_category = item['sf_category']
                        prev_sf_block_name = block_name

                if 'new_block_name' not in item:
                    item['new_block_name'] = block_name

            ordered_block_names = []
            for item in category_order:
                block_name = item['new_block_name'] if 'new_block_name' in item else item['block_name']
                if block_name not in ordered_block_names:
                    ordered_block_names.append(block_name)

            _category_order = []
            for _block_name in ordered_block_names:
                count = 0
                for item in category_order:
                    block_name = item['new_block_name'] if 'new_block_name' in item else item['block_name']
                    if block_name == _block_name:
                        _category_order.append(item)
                        count += 1
                if count == 1:
                    last_item = _category_order[-1]
                    if not last_item['sf_category_flag'] and last_item['ordered']:
                        last_item['missing_sf_category'] = True
                        last_item['new_block_name'] = last_item['block_name']

            category_order = _category_order

            _sf_category = ''
            cur_list_id = 1

            reserved_block_names = []
            sf = None
            for item in category_order:
                block_name = item['block_name']
                category = item['category']
                sf_category = item['sf_category']
                ext = item['block_name_ext']

                if item['sf_category_flag'] or ('missing_sf_category' in item and item['missing_sf_category']):
                    if sf_category not in sf_category_counter:
                        cur_list_id = 1
                    else:
                        cur_list_id = sf_category_counter[sf_category] + 1
                elif sf_category != _sf_category:
                    cur_list_id = 1

                _sf_category = sf_category

                sf_tag_prefix = next(v['Tag category'] for k, v in self.schema.items()
                                     if v['SFCategory'] == sf_category and v['Tag field'] == 'Sf_category')

                sf_category_counter[sf_category] = cur_list_id
                _cur_list_id = str(cur_list_id)

                if 'missing_sf_category' in item and item['missing_sf_category']:
                    new_block_name = item['new_block_name']

                    if new_block_name in reserved_block_names:
                        continue

                    if sf is not None:
                        strData.add_saveframe(sf)

                    sf = pynmrstar.Saveframe.from_scratch(new_block_name)
                    reserved_block_names.append(new_block_name)
                    sf.set_tag_prefix(sf_tag_prefix)
                    sf.add_tag('Sf_category', sf_category)
                    sf.add_tag('Sf_framecode', new_block_name)
                    sf.add_tag('Entry_ID', _entry_id)

                    if sf_category != item['super_category']:
                        sf.add_tag('ID', _cur_list_id)

                elif item['sf_category_flag']:
                    new_block_name = item['new_block_name'] if 'new_block_name' in item else block_name

                    if new_block_name in reserved_block_names:
                        continue

                    if sf is not None:
                        strData.add_saveframe(sf)

                    sf = pynmrstar.Saveframe.from_scratch(new_block_name)
                    reserved_block_names.append(new_block_name)
                    sf.set_tag_prefix(category)

                    dict_list = cifObj.GetDataBlock(block_name, ext)
                    itVals = next(v for k, v in dict_list.items() if k == category)

                    sf.add_tag('Sf_category', sf_category)
                    sf.add_tag('Sf_framecode', new_block_name)
                    if sf_category != 'entry_information':
                        sf.add_tag('Entry_ID', _entry_id)
                        sf.add_tag('ID', _cur_list_id)
                    else:
                        sf.add_tag('ID', _entry_id)

                    for _item, value in zip(itVals['Items'], itVals['Values'][0]):
                        if _item not in ('Sf_category', 'Sf_framecode', 'Entry_ID', 'ID'):
                            sf.add_tag(_item, value)

                if not item['sf_category_flag']:
                    lp = pynmrstar.Loop.from_scratch(category)

                    dict_list = cifObj.GetDataBlock(block_name, ext)
                    itVals = next(v for k, v in dict_list.items() if k == category)

                    list_id_idx = -1
                    entry_id_idx = -1

                    for idx, _item in enumerate(itVals['Items']):
                        lp.add_tag(_item)
                        if sf_category != item['super_category']:
                            tag = '_' + category + '.' + _item
                            tag = tag.lower()
                            if tag in self.schema:
                                sdict = {k: v for k, v in self.schema[tag].items() if v not in emptyValue}
                                if 'Parent tag' in sdict and sdict['Parent tag'] == '_' + sf_tag_prefix + '.ID':
                                    list_id_idx = idx
                        if _item == 'Entry_ID':
                            entry_id_idx = idx

                    for idx, _value in enumerate(itVals['Values']):
                        if list_id_idx != -1:
                            _value[list_id_idx] = _cur_list_id
                        if entry_id_idx != -1:
                            _value[entry_id_idx] = _entry_id
                        lp.add_data(_value)

                    if sf is not None:
                        sf.add_loop(lp)

            if sf is not None:
                strData.add_saveframe(sf)

            self.normalize(strData)

            if __pynmrstar_v3__:
                strData.write_to_file(strPath, show_comments=False, skip_empty_loops=True, skip_empty_tags=False)
            else:
                strData.write_to_file(strPath)

            return True

        except ValueError as e:

            sf_anonymous_pattern = re.compile(r'\s*save_\S+\s*')
            save_pattern = re.compile(r'\s*save_\s*')

            split_ext = os.path.splitext(cifPath)
            _cifPath = split_ext[0] + '-corrected' + ('' if len(split_ext) == 1 else split_ext[1])

            changed = False

            with open(cifPath, 'r') as ifh, \
                    open(_cifPath, 'w') as ofh:
                for line in ifh:
                    if sf_anonymous_pattern.match(line) or save_pattern.match(line):
                        ofh.write('#' + line)
                        changed = True
                    else:
                        ofh.write(line)

            if changed and maxRepeat > 0:
                return self.convert(_cifPath, strPath, datablockName, maxRepeat=maxRepeat - 1)

            try:
                os.remove(_cifPath)
            except OSError:
                pass

            self.__lfh.write(f"+ERROR- CifToNmrStar.convert() {str(e)}\n")

            return False

        except Exception as e:
            self.__lfh.write(f"+ERROR- CifToNmrStar.convert() {str(e)}\n")

            return False

    def set_entry_id(self, strData, entryId):
        """ Set entry ID without changing datablock name.
            @see: pynmrstar.entry
        """

        if strData is None:
            return

        entryId = entryId.strip().replace(' ', '_')  # DAOTHER-9511: replace white space in a datablock name to underscore

        if isinstance(strData, pynmrstar.Entry):

            for sf in strData.frame_list:
                filled = False

                for tag in sf.tags:
                    fqtn = (sf.tag_prefix + '.' + tag[0]).lower()

                    try:
                        if self.schema[fqtn]['entryIdFlg'] == 'Y':
                            tag[1] = entryId
                            filled = True
                            break
                    except KeyError:
                        pass

                if not filled:
                    entry_id_tag = 'ID' if sf.category == 'entry_information' else 'Entry_ID'

                    fqtn = (sf.tag_prefix + '.' + entry_id_tag).lower()

                    try:
                        if self.schema[fqtn]['entryIdFlg'] == 'Y':
                            sf.add_tag(entry_id_tag, entryId)
                    except KeyError:
                        pass

                for lp in sf.loops:
                    filled = False

                    for tag in lp.tags:
                        fqtn = (lp.category + '.' + tag).lower()

                        try:
                            if self.schema[fqtn]['entryIdFlg'] == 'Y' or tag == 'Entry_ID':
                                lp[tag] = [entryId] * len(lp[tag])
                                filled = True
                                break
                        except KeyError:
                            pass

                    if not filled:
                        entry_id_tag = lp.category + '.Entry_ID'

                        lp.add_tag(entry_id_tag)

                        for row in lp:
                            row.append(entryId)

        elif isinstance(strData, pynmrstar.Saveframe):

            sf = strData

            filled = False

            for tag in sf.tags:
                fqtn = (sf.tag_prefix + '.' + tag[0]).lower()

                try:
                    if self.schema[fqtn]['entryIdFlg'] == 'Y':
                        tag[1] = entryId
                        filled = True
                        break
                except KeyError:
                    pass

            if not filled:
                entry_id_tag = 'ID' if sf.category == 'entry_information' else 'Entry_ID'

                fqtn = (sf.tag_prefix + '.' + entry_id_tag).lower()

                try:
                    if self.schema[fqtn]['entryIdFlg'] == 'Y':
                        sf.add_tag(entry_id_tag, entryId)
                except KeyError:
                    pass

            for lp in sf.loops:
                filled = False

                for tag in lp.tags:
                    fqtn = (lp.category + '.' + tag).lower()

                    try:
                        if self.schema[fqtn]['entryIdFlg'] == 'Y' or tag == 'Entry_ID':
                            lp[tag] = [entryId] * len(lp[tag])
                            filled = True
                            break
                    except KeyError:
                        pass

                if not filled:
                    entry_id_tag = lp.category + '.Entry_ID'

                    lp.add_tag(entry_id_tag)

                    for row in lp:
                        row.append(entryId)

        elif isinstance(strData, pynmrstar.Loop):

            lp = strData

            filled = False

            for tag in lp.tags:
                fqtn = (lp.category + '.' + tag).lower()

                try:
                    if self.schema[fqtn]['entryIdFlg'] == 'Y' or tag == 'Entry_ID':
                        lp[tag] = [entryId] * len(lp[tag])
                        filled = True
                        break
                except KeyError:
                    pass

            if not filled:
                entry_id_tag = lp.category + '.Entry_ID'

                lp.add_tag(entry_id_tag)

                for row in lp:
                    row.append(entryId)

    def set_local_sf_id(self, strData, listId):
        """ Set list ID for a given saveframe or loop.
        """

        if strData is None:
            return

        if isinstance(strData, pynmrstar.Saveframe):

            sf = strData

            for tag in sf.tags:
                fqtn = (sf.tag_prefix + '.' + tag[0]).lower()

                try:
                    if self.schema[fqtn]['lclSfIdFlg'] == 'Y':
                        tag[1] = listId
                        break
                except KeyError:
                    pass

            for lp in sf.loops:

                for tag in lp.tags:
                    fqtn = (lp.category + '.' + tag).lower()

                    try:
                        if self.schema[fqtn]['lclSfIdFlg'] == 'Y':
                            lp[tag] = [listId] * len(lp[tag])
                            break
                    except KeyError:
                        pass

        elif isinstance(strData, pynmrstar.Loop):

            lp = strData

            for tag in lp.tags:
                fqtn = (lp.category + '.' + tag).lower()

                try:
                    if self.schema[fqtn]['lclSfIdFlg'] == 'Y':
                        lp[tag] = [listId] * len(lp[tag])
                        break
                except KeyError:
                    pass

    def normalize(self, strData):
        """ Wrapper function of normalize_str() and normalize_nef().
        """

        try:
            sf = strData.frame_list[0]
            if sf.category.startswith('nef'):
                return self.normalize_nef(strData)
            return self.normalize_str(strData)
        except (IndexError, AttributeError):
            return strData

    def normalize_str(self, strData):
        """ Sort saveframes, loops, and tags according to NMR-STAR schema.
            @see: pynmrstar.entry.normalize
        """

        def sf_key(sf):
            """ Helper function to sort the saveframes.
            Returns (category order, saveframe order) """

            # If not a real category, generate an artificial but stable order > the real saveframes
            try:
                category_order = self.category_order.index(sf.tag_prefix)
            except (ValueError, KeyError):
                if sf.category is None:
                    category_order = float('infinity')
                else:
                    category_order = len(self.category_order) + abs(int(hashlib.sha1(str(sf.category).encode()).hexdigest(), 16))

            # See if there is an ID tag, and it is a number
            saveframe_id = float('infinity')
            try:
                saveframe_id = int(sf.get_tag("ID")[0])
            except (ValueError, KeyError, IndexError, TypeError):
                # Either there is no ID, or it is not a number. By default it will sort at the end of saveframes of its
                # category. Note that the entry_information ID tag has a different meaning, but since there should
                # only ever be one saveframe of that category, the sort order for it can be any value.
                pass

            return category_order, saveframe_id

        def lp_key(lp):
            """ Helper function to sort the loops.
            Returns (category order) """

            try:
                category_order = self.category_order.index(lp.category)
            except (ValueError, KeyError):
                category_order = float('infinity')

            return category_order

        try:
            strData.frame_list.sort(key=sf_key)
        except Exception as e:
            self.__lfh.write(f"+ERROR- CifToNmrStar.normalize() {str(e)}\n")

        for sf in strData.frame_list:
            sf.sort_tags()
            if len(sf.loops) > 1:
                sf.loops.sort(key=lp_key)
            # Iterate through the loops
            for lp in sf:
                lp.sort_tags()

        return strData

    def normalize_nef(self, strData):
        """ Sort saveframes of NEF.
        """

        def sf_key(sf):
            """ Helper function to sort the saveframes.
            Returns (saveframe order) """

            try:
                category_order = self.category_order_nef.index(sf.tag_prefix)
            except (ValueError, KeyError):
                if sf.category is None:
                    category_order = float('infinity')
                else:
                    category_order = len(self.category_order_nef) + abs(int(hashlib.sha1(str(sf.category).encode()).hexdigest(), 16))

            return category_order

        try:
            strData.frame_list.sort(key=sf_key)
        except Exception as e:
            self.__lfh.write(f"+ERROR- CifToNmrStar.normalize_nef() {str(e)}\n")

        return strData



##
# XplorMRReader.py
#
# Update:
##
""" A collection of classes for parsing XPLOR-NIH MR files.
"""
import sys
import os

from antlr4 import InputStream, CommonTokenStream, ParseTreeWalker, PredictionMode

try:
    from wwpdb.utils.nmr.mr.LexerErrorListener import LexerErrorListener
    from wwpdb.utils.nmr.mr.ParserErrorListener import ParserErrorListener
    from wwpdb.utils.nmr.mr.XplorMRLexer import XplorMRLexer
    from wwpdb.utils.nmr.mr.XplorMRParser import XplorMRParser
    from wwpdb.utils.nmr.mr.XplorMRParserListener import XplorMRParserListener
    from wwpdb.utils.nmr.mr.ParserListenerUtil import (coordAssemblyChecker,
                                                       MAX_ERROR_REPORT,
                                                       REPRESENTATIVE_MODEL_ID,
                                                       REPRESENTATIVE_ALT_ID)
    from wwpdb.utils.nmr.io.CifReader import CifReader
    from wwpdb.utils.nmr.ChemCompUtil import ChemCompUtil
    from wwpdb.utils.nmr.BMRBChemShiftStat import BMRBChemShiftStat
    from wwpdb.utils.nmr.NEFTranslator.NEFTranslator import NEFTranslator
except ImportError:
    from nmr.mr.LexerErrorListener import LexerErrorListener
    from nmr.mr.ParserErrorListener import ParserErrorListener
    from nmr.mr.XplorMRLexer import XplorMRLexer
    from nmr.mr.XplorMRParser import XplorMRParser
    from nmr.mr.XplorMRParserListener import XplorMRParserListener
    from nmr.mr.ParserListenerUtil import (coordAssemblyChecker,
                                           MAX_ERROR_REPORT,
                                           REPRESENTATIVE_MODEL_ID,
                                           REPRESENTATIVE_ALT_ID)
    from nmr.io.CifReader import CifReader
    from nmr.ChemCompUtil import ChemCompUtil
    from nmr.BMRBChemShiftStat import BMRBChemShiftStat
    from nmr.NEFTranslator.NEFTranslator import NEFTranslator


class XplorMRReader:
    """ Accessor methods for parsing XPLOR-NIH MR files.
    """

    def __init__(self, verbose=True, log=sys.stdout,
                 representativeModelId=REPRESENTATIVE_MODEL_ID,
                 representativeAltId=REPRESENTATIVE_ALT_ID,
                 mrAtomNameMapping=None,
                 cR=None, caC=None, ccU=None, csStat=None, nefT=None,
                 reasons=None):
        self.__verbose = verbose
        self.__lfh = log
        self.__debug = False
        self.__remediate = False
        self.__sll_pred = False

        self.__maxLexerErrorReport = MAX_ERROR_REPORT
        self.__maxParserErrorReport = MAX_ERROR_REPORT

        self.__representativeModelId = representativeModelId
        self.__representativeAltId = representativeAltId
        self.__mrAtomNameMapping = mrAtomNameMapping

        if cR is not None and caC is None:
            caC = coordAssemblyChecker(verbose, log, representativeModelId, representativeAltId,
                                       cR, None, None, fullCheck=False)

        self.__cR = cR
        self.__caC = caC

        # CCD accessing utility
        self.__ccU = ChemCompUtil(verbose, log) if ccU is None else ccU

        # BMRB chemical shift statistics
        self.__csStat = BMRBChemShiftStat(verbose, log, self.__ccU) if csStat is None else csStat

        # NEFTranslator
        self.__nefT = NEFTranslator(verbose, log, self.__ccU, self.__csStat) if nefT is None else nefT
        if nefT is None:
            self.__nefT.set_remediation_mode(True)

        # reasons for re-parsing request from the previous trial
        self.__reasons = reasons

    def setDebugMode(self, debug):
        self.__debug = debug

    def setRemediateMode(self, remediate):
        self.__remediate = remediate

    def setLexerMaxErrorReport(self, maxErrReport):
        self.__maxLexerErrorReport = maxErrReport

    def setParserMaxErrorReport(self, maxErrReport):
        self.__maxParserErrorReport = maxErrReport

    def setSllPredMode(self, ssl_pred):
        self.__sll_pred = ssl_pred

    def parse(self, mrFilePath, cifFilePath=None, isFilePath=True,
              createSfDict=False, originalFileName=None, listIdCounter=None, entryId=None):
        """ Parse XPLOR-NIH MR file.
            @return: XplorMRParserListener for success or None otherwise, ParserErrorListener, LexerErrorListener.
        """

        ifh = None

        try:

            if isFilePath:
                mrString = None

                if not os.access(mrFilePath, os.R_OK):
                    if self.__verbose:
                        self.__lfh.write(f"XplorMRReader.parse() {mrFilePath} is not accessible.\n")
                    return None, None, None

                ifh = open(mrFilePath, 'r')  # pylint: disable=consider-using-with
                input = InputStream(ifh.read())

            else:
                mrFilePath, mrString = None, mrFilePath

                if mrString is None or len(mrString) == 0:
                    if self.__verbose:
                        self.__lfh.write("XplorMRReader.parse() Empty string.\n")
                    return None, None, None

                input = InputStream(mrString)

            if cifFilePath is not None:
                if not os.access(cifFilePath, os.R_OK):
                    if self.__verbose:
                        self.__lfh.write(f"XplorMRReader.parse() {cifFilePath} is not accessible.\n")
                    return None, None, None

                if self.__cR is None:
                    self.__cR = CifReader(self.__verbose, self.__lfh)
                    if not self.__cR.parse(cifFilePath):
                        return None, None, None

            lexer = XplorMRLexer(input)
            lexer.removeErrorListeners()

            lexer_error_listener = LexerErrorListener(mrFilePath, inputString=mrString, maxErrorReport=self.__maxLexerErrorReport)
            lexer.addErrorListener(lexer_error_listener)

            messageList = lexer_error_listener.getMessageList()

            if messageList is not None and self.__verbose:
                for description in messageList:
                    self.__lfh.write(f"[Syntax error] line {description['line_number']}:{description['column_position']} {description['message']}\n")
                    if 'input' in description:
                        self.__lfh.write(f"{description['input']}\n")
                        self.__lfh.write(f"{description['marker']}\n")

            stream = CommonTokenStream(lexer)
            parser = XplorMRParser(stream)
            if not isFilePath or self.__sll_pred:
                parser._interp.predictionMode = PredictionMode.SLL  # pylint: disable=protected-access
            parser.removeErrorListeners()
            parser_error_listener = ParserErrorListener(mrFilePath, inputString=mrString, maxErrorReport=self.__maxParserErrorReport)
            parser.addErrorListener(parser_error_listener)
            tree = parser.xplor_nih_mr()

            walker = ParseTreeWalker()
            listener = XplorMRParserListener(self.__verbose, self.__lfh,
                                             self.__representativeModelId,
                                             self.__representativeAltId,
                                             self.__mrAtomNameMapping,
                                             self.__cR, self.__caC,
                                             self.__ccU, self.__csStat, self.__nefT,
                                             self.__reasons)
            listener.setDebugMode(self.__debug)
            listener.setRemediateMode(self.__remediate)
            listener.createSfDict(createSfDict)
            if createSfDict:
                if originalFileName is not None:
                    listener.setOriginaFileName(originalFileName)
                if listIdCounter is not None:
                    listener.setListIdCounter(listIdCounter)
                if entryId is not None:
                    listener.setEntryId(entryId)
            walker.walk(listener, tree)

            messageList = parser_error_listener.getMessageList()

            if messageList is not None and self.__verbose:
                for description in messageList:
                    self.__lfh.write(f"[Syntax error] line {description['line_number']}:{description['column_position']} {description['message']}\n")
                    if 'input' in description:
                        self.__lfh.write(f"{description['input']}\n")
                        self.__lfh.write(f"{description['marker']}\n")

            if self.__verbose:
                if listener.warningMessage is not None and len(listener.warningMessage) > 0:
                    print('\n'.join(listener.warningMessage))
                if isFilePath:
                    print(listener.getContentSubtype())

            return listener, parser_error_listener, lexer_error_listener

        except IOError as e:
            if self.__verbose:
                self.__lfh.write(f"+XplorMRReader.parse() ++ Error - {str(e)}\n")
            return None, None, None
            # pylint: disable=unreachable
            """ debug code
        except Exception as e:
            if self.__verbose and isFilePath:
                self.__lfh.write(f"+XplorMRReader.parse() ++ Error - {mrFilePath!r} - {str(e)}\n")
            return None, None, None
            """
        finally:
            if isFilePath and ifh is not None:
                ifh.close()


if __name__ == "__main__":
    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/5t1n/PCS_HN.tbl',
                     '../../tests-nmr/mock-data-remediation/5t1n/5t1n.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/2m6z/2m6z-corrected.mr.1',
                     '../../tests-nmr/mock-data-remediation/2m6z/2m6z.cif')
    print(reader_listener.getReasonsForReparsing())

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/2mlz/2mlz-corrected.mr',
                     '../../tests-nmr/mock-data-remediation/2mlz/2mlz.cif')
    print(reader_listener.getReasonsForReparsing())

    reader = XplorMRReader(False)
    reader.setDebugMode(False)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/4b1q/4b1q-trimmed.mr',
                     '../../tests-nmr/mock-data-remediation/4b1q/4b1q.cif')
    print(reader_listener.getReasonsForReparsing())
    reader = XplorMRReader(True, reasons=reader_listener.getReasonsForReparsing())
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/4b1q/4b1q-trimmed.mr',
                 '../../tests-nmr/mock-data-remediation/4b1q/4b1q.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2kf4/2kf4-trimmed.mr',
                 '../../tests-nmr/mock-data-remediation/2kf4/2kf4.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/1iym/1iym-trimmed.mr',
                     '../../tests-nmr/mock-data-remediation/1iym/1iym.cif')
    print(reader_listener.getReasonsForReparsing())

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/5oay/cluster_noe2.tbl',
                     '../../tests-nmr/mock-data-remediation/5oay/5oay.cif')
    print(reader_listener.getReasonsForReparsing())
    reader = XplorMRReader(True, reasons=reader_listener.getReasonsForReparsing())
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/5oay/cluster_noe2.tbl',
                 '../../tests-nmr/mock-data-remediation/5oay/5oay.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/6uz4/AGL55_K2_NOE.tbl',
                     '../../tests-nmr/mock-data-remediation/6uz4/6uz4.cif')
    print(reader_listener.getReasonsForReparsing())

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/7t1o/noe.tbl-corrected',
                     '../../tests-nmr/mock-data-remediation/7t1o/7t1o.cif')
    print(reader_listener.getReasonsForReparsing())

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/2ll4/2ll4-trimmed.mr',
                     '../../tests-nmr/mock-data-remediation/2ll4/2ll4.cif')
    print(reader_listener.getReasonsForReparsing())

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/5j2w/TARpRAM_pf1.txt-corrected',
                     '../../tests-nmr/mock-data-remediation/5j2w/5j2w.cif')
    print(reader_listener.getReasonsForReparsing())

    reader = XplorMRReader(False)
    reader.setDebugMode(False)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/2rsy/2rsy-corrected-div_dst-div_dst.mr',
                     '../../tests-nmr/mock-data-remediation/2rsy/2rsy.cif')
    print(reader_listener.getReasonsForReparsing())

    reader = XplorMRReader(False)
    reader.setDebugMode(False)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/2mpg/2mpg-trimmed.mr',
                     '../../tests-nmr/mock-data-remediation/2mpg/2mpg.cif')
    print(reader_listener.getReasonsForReparsing())
    reader = XplorMRReader(True, reasons=reader_listener.getReasonsForReparsing())
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2mpg/2mpg-trimmed.mr',
                 '../../tests-nmr/mock-data-remediation/2mpg/2mpg.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/2kxa/2kxa-trimmed.mr',
                     '../../tests-nmr/mock-data-remediation/2kxa/2kxa.cif')
    print(reader_listener.getReasonsForReparsing())

    reader = XplorMRReader(False)
    reader.setDebugMode(False)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/7t03/HB1_noe.tbl',
                     '../../tests-nmr/mock-data-remediation/7t03/7t03.cif')
    print(reader_listener.getReasonsForReparsing())

    reader = XplorMRReader(False)
    reader.setDebugMode(False)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/2lzs/2lzs-trimmed.mr',
                     '../../tests-nmr/mock-data-remediation/2lzs/2lzs.cif')
    print(reader_listener.getReasonsForReparsing())

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/2m3t/test.mr',
                     '../../tests-nmr/mock-data-remediation/2m3t/2m3t.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/5j0m/TARpRAM_pf1.txt-corrected',
                     '../../tests-nmr/mock-data-remediation/5j0m/5j0m.cif')

    reader = XplorMRReader(False)
    reader.setDebugMode(False)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/2n1e/2n1e-corrected-div_src.mr',
                     '../../tests-nmr/mock-data-remediation/2n1e/2n1e.cif')
    print(reader_listener.getReasonsForReparsing())
    reader = XplorMRReader(True, reasons=reader_listener.getReasonsForReparsing())
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2n1e/2n1e-corrected-div_src.mr',
                 '../../tests-nmr/mock-data-remediation/2n1e/2n1e.cif')

    reader = XplorMRReader(False)
    reader.setDebugMode(False)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/5w3n/fus_CC_PITHIRDSCT.tbl',
                     '../../tests-nmr/mock-data-remediation/5w3n/5w3n.cif')
    print(reader_listener.getReasonsForReparsing())
    reader = XplorMRReader(True, reasons=reader_listener.getReasonsForReparsing())
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/5w3n/fus_CC_PITHIRDSCT.tbl',
                 '../../tests-nmr/mock-data-remediation/5w3n/5w3n.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2mnh/2mnh-trimmed-div_src.mr',
                 '../../tests-nmr/mock-data-remediation/2mnh/2mnh.cif')

    reader = XplorMRReader(False)
    reader.setDebugMode(False)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/7m3u/PDP-24_restraints.tbl',
                     '../../tests-nmr/mock-data-remediation/7m3u/7m3u.cif')
    print(reader_listener.getReasonsForReparsing())

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2n7j/2n7j-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/2n7j/2n7j.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2n05/2n05-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/2n05/2n05.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/5ggm/dis.txt',
                 '../../tests-nmr/mock-data-remediation/5ggm/5ggm.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/5y95/5y95-trimmed.mr',
                 '../../tests-nmr/mock-data-remediation/5y95/5y95.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2lai/2lai-trimmed.mr',
                 '../../tests-nmr/mock-data-remediation/2lai/2lai.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2ruj/test.mr',
                 '../../tests-nmr/mock-data-remediation/2ruj/2ruj.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/2mus/2mus-corrected.mr',
                     '../../tests-nmr/mock-data-remediation/2mus/2mus.cif')
    reader = XplorMRReader(True, reasons=reader_listener.getReasonsForReparsing())
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2mus/2mus-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/2mus/2mus.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/6qeu/disu.tbl',
                 '../../tests-nmr/mock-data-remediation/6qeu/6qeu.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/5kmz/TPK_restraints.txt',
                 '../../tests-nmr/mock-data-remediation/5kmz/5kmz.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/6bvy/test.mr',
                 '../../tests-nmr/mock-data-remediation/6bvy/6bvy.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2lck/2lck-trimmed.mr',
                 '../../tests-nmr/mock-data-remediation/2lck/2lck.cif')

    reader = XplorMRReader(False)
    reader.setDebugMode(False)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/2n1a/2n1a-corrected.mr',
                     '../../tests-nmr/mock-data-remediation/2n1a/2n1a.cif')
    reader = XplorMRReader(True, reasons=reader_listener.getReasonsForReparsing())
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2n1a/2n1a-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/2n1a/2n1a.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2mjf/2mjf-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/2mjf/2mjf.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2muk/2muk-trimmed.mr',
                 '../../tests-nmr/mock-data-remediation/2muk/2muk.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2kny/2kny-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/2kny/2kny.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2ltj/2ltj-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/2ltj/2ltj.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2n24/2n24-trimmed.mr',
                 '../../tests-nmr/mock-data-remediation/2n24/2n24.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/6feg/03_NOEs_unambig.tbl',
                     '../../tests-nmr/mock-data-remediation/6feg/6feg.cif')
    reader = XplorMRReader(True, reasons=reader_listener.getReasonsForReparsing())
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/6feg/03_NOEs_unambig.tbl',
                 '../../tests-nmr/mock-data-remediation/6feg/6feg.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/6feg/01_Dihedral_XPLOR.tbl-corrected',
                     '../../tests-nmr/mock-data-remediation/6feg/6feg.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/6feg/02_RDCs_XPLOR.tbl-corrected',
                     '../../tests-nmr/mock-data-remediation/6feg/6feg.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2mqi/2mqi-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/2mqi/2mqi.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/7ui5/jna_coup_dimer_2.tbl-corrected',
                 '../../tests-nmr/mock-data-remediation/7ui5/7ui5.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2ml9/2ml9-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/2ml9/2ml9.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/3zfj/3zfj-trimmed.mr',
                     '../../tests-nmr/mock-data-remediation/3zfj/3zfj.cif')
    reader = XplorMRReader(True, reasons=reader_listener.getReasonsForReparsing())
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/3zfj/3zfj-trimmed.mr',
                 '../../tests-nmr/mock-data-remediation/3zfj/3zfj.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/2b87/2b87-corrected.mr',
                     '../../tests-nmr/mock-data-remediation/2b87/2b87.cif')
    reader = XplorMRReader(True, reasons=reader_listener.getReasonsForReparsing())
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2b87/2b87-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/2b87/2b87.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/2mj5/2mj5-corrected.mr',
                     '../../tests-nmr/mock-data-remediation/2mj5/2mj5.cif')
    reader = XplorMRReader(True, reasons=reader_listener.getReasonsForReparsing())
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2mj5/2mj5-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/2mj5/2mj5.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/5jyv/DN-HN_ThCikA_gel_complex_complete.tbl',
                     '../../tests-nmr/mock-data-remediation/5jyv/5jyv.cif')
    reader = XplorMRReader(True, reasons=reader_listener.getReasonsForReparsing())
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/5jyv/DN-HN_ThCikA_gel_complex_complete.tbl',
                 '../../tests-nmr/mock-data-remediation/5jyv/5jyv.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/5jyv/c13noe-square_cika.tbl',
                     '../../tests-nmr/mock-data-remediation/5jyv/5jyv.cif')
    reader = XplorMRReader(True, reasons=reader_listener.getReasonsForReparsing())
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/5jyv/c13noe-square_cika.tbl',
                 '../../tests-nmr/mock-data-remediation/5jyv/5jyv.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-combine-at-upload/bmr21102/sms21102/AmBErg_CC_restraints_PAR12p6ms.tbl',
                 '../../tests-nmr/mock-data-combine-at-upload/bmr21102/sms21102/sms21102.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-combine-at-upload/bmr21100/sms21100/Ambig.tbl',
                 '../../tests-nmr/mock-data-combine-at-upload/bmr21100/sms21100/sms21100.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-daother-8580/newupload_2/D_1000274286_mr_P3.xplor-nih.V1',
                     '../../tests-nmr/mock-data-daother-8580/newupload_2/D_1000274286_model_P1.cif.V1')
    reader = XplorMRReader(True, reasons=reader_listener.getReasonsForReparsing())
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-daother-8580/newupload_2/D_1000274286_mr_P3.xplor-nih.V1',
                 '../../tests-nmr/mock-data-daother-8580/newupload_2/D_1000274286_model_P1.cif.V1')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2mjb/2mjb-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/2mjb/2mjb.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2l94/2l94-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/2l94/2l94.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/5t1o/PCS_HN_EIN.tbl-corrected',
                 '../../tests-nmr/mock-data-remediation/5t1o/5t1o.cif')

    reader = XplorMRReader(False)
    reader.setDebugMode(False)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/2n8l/2n8l-corrected.mr',
                     '../../tests-nmr/mock-data-remediation/2n8l/2n8l.cif')
    reader = XplorMRReader(True, reasons=reader_listener.getReasonsForReparsing())
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2n8l/2n8l-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/2n8l/2n8l.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/6th8/allnoe.tbl-corrected',
                 '../../tests-nmr/mock-data-remediation/6th8/6th8.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2mjk/2mjk-trimmed.mr',
                 '../../tests-nmr/mock-data-remediation/2mjk/2mjk.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/6cch/6cch-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/6cch/6cch.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2lrk/2lrk-corrected-div_src.mr',
                 '../../tests-nmr/mock-data-remediation/2lrk/2lrk.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2mk3/2mk3-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/2mk3/2mk3.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/5xbo/Tb_UBA1_PCS_correct_50p.tbl-corrected',
                 '../../tests-nmr/mock-data-remediation/5xbo/5xbo.cif')

    reader = XplorMRReader(False)
    reader.setDebugMode(False)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/2k31/2k31-corrected.mr',
                     '../../tests-nmr/mock-data-remediation/2k31/2k31.cif')

    reader = XplorMRReader(True, reasons=reader_listener.getReasonsForReparsing())
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2k31/2k31-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/2k31/2k31.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2lrk/2lrk-corrected-div_src.mr',
                 '../../tests-nmr/mock-data-remediation/2lrk/2lrk.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/1iio/1iio-trimmed.mr',
                     '../../tests-nmr/mock-data-remediation/1iio/1iio.cif')
    reader = XplorMRReader(True, reasons=reader_listener.getReasonsForReparsing())
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/1iio/1iio-trimmed.mr',
                 '../../tests-nmr/mock-data-remediation/1iio/1iio.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2m3g/2m3g-trimmed.mr',
                 '../../tests-nmr/mock-data-remediation/2m3g/2m3g.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/7n7e/distance.xplor',
                 '../../tests-nmr/mock-data-remediation/7n7e/7n7e.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader_listener, _, _ =\
        reader.parse('../../tests-nmr/mock-data-remediation/2ju5/2ju5-corrected-div_src.mr',
                     '../../tests-nmr/mock-data-remediation/2ju5/2ju5.cif')
    reader = XplorMRReader(True, reasons=reader_listener.getReasonsForReparsing())
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2ju5/2ju5-corrected-div_src.mr',
                 '../../tests-nmr/mock-data-remediation/2ju5/2ju5.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/6x63/deposit_All_Unamb_MD_xplorFormat_fixNomen.txt',
                 '../../tests-nmr/mock-data-remediation/6x63/6x63.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2laj/2laj-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/2laj/2laj.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/5kqj/ANT2_hbond_restraint.tbl',
                 '../../tests-nmr/mock-data-remediation/5kqj/5kqj.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2l1v/2l1v-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/2l1v/2l1v.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2mcc/2mcc-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/2mcc/2mcc.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/6d53/dangleT.tbl-corrected',
                 '../../tests-nmr/mock-data-remediation/6d53/6d53.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2k2a/2k2a-corrected-div_src.mr',
                 '../../tests-nmr/mock-data-remediation/2k2a/2k2a.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/5w3n/fus_CC_RDC.tbl-corrected',
                 '../../tests-nmr/mock-data-remediation/5w3n/5w3n.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/5w3n/fus_CO_RDC.tbl-corrected',
                 '../../tests-nmr/mock-data-remediation/5w3n/5w3n.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/6var/HBVe_sPRE.tab-corrected',
                 '../../tests-nmr/mock-data-remediation/6var/6var.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/6l92/G4II_planarity.txt',
                 '../../tests-nmr/mock-data-remediation/6l92/6l92.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/5t1o/RDC_NPr.tbl-corrected',
                 '../../tests-nmr/mock-data-remediation/5t1o/5t1o.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/6ud0/pcs_hn_dist_tm_tsg_tsap.tbl-corrected',
                 '../../tests-nmr/mock-data-remediation/6ud0/6ud0.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2l3b/cacb_shift.mr',
                 '../../tests-nmr/mock-data-remediation/2l3b/2l3b.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2kw8/2kw8-trimmed-div_dst-div_src.mr',
                 '../../tests-nmr/mock-data-remediation/2kw8/2kw8.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2bjc/test_no_eff.mr',
                 '../../tests-nmr/mock-data-remediation/2bjc/2bjc.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2bgo/test.mr',
                 '../../tests-nmr/mock-data-remediation/2bgo/2bgo.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/5ks5/vean.test',
                 '../../tests-nmr/mock-data-remediation/5ks5/5ks5.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2bjc/test.mr',
                 '../../tests-nmr/mock-data-remediation/2bjc/2bjc.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/4by9/4by9-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/4by9/4by9.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-daother-7871/wsh2090_ambig_rev.tbl',
                 '../../tests-nmr/mock-data-daother-7871/D_800473_model_P1.cif.V3')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2n55/test.mr',
                 '../../tests-nmr/mock-data-remediation/2n55/2n55.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2m54/2m54-trimmed.mr',
                 '../../tests-nmr/mock-data-remediation/2m54/2m54.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/2n21/2n21-corrected.mr',
                 '../../tests-nmr/mock-data-remediation/2n21/2n21.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/6p6c/pea-15erk2c13shifts_ded.tbl',
                 '../../tests-nmr/mock-data-remediation/6p6c/6p6c.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-remediation/6jcd/submit-plane.tbl',
                 '../../tests-nmr/mock-data-remediation/6jcd/6jcd.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-daother-7690/D_1300020446_mr-upload_P5.xplor-nih.V1',
                 '../../tests-nmr/mock-data-daother-7690/D_1300020446_model-release_P1.cif.V2')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-daother-7690/D_1300020646_mr-upload_P1.xplor-nih.V2',
                 '../../tests-nmr/mock-data-daother-7690/D_1300020646_model-annotate_P1.cif.V2')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-pdbstat/D_1000243168_mr-upload_P8.xplor-nih.V1',
                 '../../tests-nmr/mock-data-pdbstat/6pvr.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-pdbstat/D_1000243168_mr-upload_P2.xplor-nih.V1',
                 '../../tests-nmr/mock-data-pdbstat/6pvr.cif')

    reader = XplorMRReader(True)
    reader.setDebugMode(True)
    reader.parse('../../tests-nmr/mock-data-pdbstat/atom_sel_expr_example.txt',
                 '../../tests-nmr/mock-data-pdbstat/6pvr.cif')



# Generated from IsdMRParser.g4 by ANTLR 4.13.0
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,11,24,2,0,7,0,2,1,7,1,2,2,7,2,1,0,5,0,8,8,0,10,0,12,0,11,9,0,
        1,0,1,0,1,1,1,1,4,1,17,8,1,11,1,12,1,18,1,2,1,2,1,2,1,2,0,0,3,0,
        2,4,0,0,22,0,9,1,0,0,0,2,14,1,0,0,0,4,20,1,0,0,0,6,8,3,2,1,0,7,6,
        1,0,0,0,8,11,1,0,0,0,9,7,1,0,0,0,9,10,1,0,0,0,10,12,1,0,0,0,11,9,
        1,0,0,0,12,13,5,0,0,1,13,1,1,0,0,0,14,16,5,1,0,0,15,17,3,4,2,0,16,
        15,1,0,0,0,17,18,1,0,0,0,18,16,1,0,0,0,18,19,1,0,0,0,19,3,1,0,0,
        0,20,21,5,7,0,0,21,22,5,7,0,0,22,5,1,0,0,0,2,9,18
    ]

class IsdMRParser ( Parser ):

    grammarFileName = "IsdMRParser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [  ]

    symbolicNames = [ "<INVALID>", "Distance", "Integer", "Float", "SHARP_COMMENT", 
                      "EXCLM_COMMENT", "SMCLN_COMMENT", "Atom_selection", 
                      "SPACE", "ENCLOSE_COMMENT", "SECTION_COMMENT", "LINE_COMMENT" ]

    RULE_isd_mr = 0
    RULE_distance_restraints = 1
    RULE_distance_restraint = 2

    ruleNames =  [ "isd_mr", "distance_restraints", "distance_restraint" ]

    EOF = Token.EOF
    Distance=1
    Integer=2
    Float=3
    SHARP_COMMENT=4
    EXCLM_COMMENT=5
    SMCLN_COMMENT=6
    Atom_selection=7
    SPACE=8
    ENCLOSE_COMMENT=9
    SECTION_COMMENT=10
    LINE_COMMENT=11

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.0")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class Isd_mrContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(IsdMRParser.EOF, 0)

        def distance_restraints(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(IsdMRParser.Distance_restraintsContext)
            else:
                return self.getTypedRuleContext(IsdMRParser.Distance_restraintsContext,i)


        def getRuleIndex(self):
            return IsdMRParser.RULE_isd_mr

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterIsd_mr" ):
                listener.enterIsd_mr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitIsd_mr" ):
                listener.exitIsd_mr(self)




    def isd_mr(self):

        localctx = IsdMRParser.Isd_mrContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_isd_mr)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 9
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==1:
                self.state = 6
                self.distance_restraints()
                self.state = 11
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 12
            self.match(IsdMRParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Distance_restraintsContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Distance(self):
            return self.getToken(IsdMRParser.Distance, 0)

        def distance_restraint(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(IsdMRParser.Distance_restraintContext)
            else:
                return self.getTypedRuleContext(IsdMRParser.Distance_restraintContext,i)


        def getRuleIndex(self):
            return IsdMRParser.RULE_distance_restraints

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterDistance_restraints" ):
                listener.enterDistance_restraints(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitDistance_restraints" ):
                listener.exitDistance_restraints(self)




    def distance_restraints(self):

        localctx = IsdMRParser.Distance_restraintsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_distance_restraints)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 14
            self.match(IsdMRParser.Distance)
            self.state = 16 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 15
                self.distance_restraint()
                self.state = 18 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==7):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Distance_restraintContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Atom_selection(self, i:int=None):
            if i is None:
                return self.getTokens(IsdMRParser.Atom_selection)
            else:
                return self.getToken(IsdMRParser.Atom_selection, i)

        def getRuleIndex(self):
            return IsdMRParser.RULE_distance_restraint

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterDistance_restraint" ):
                listener.enterDistance_restraint(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitDistance_restraint" ):
                listener.exitDistance_restraint(self)




    def distance_restraint(self):

        localctx = IsdMRParser.Distance_restraintContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_distance_restraint)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 20
            self.match(IsdMRParser.Atom_selection)
            self.state = 21
            self.match(IsdMRParser.Atom_selection)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx





